@@ -30,8 +30,7 @@ except ImportError:
     pass
 
 from qiskit import QiskitError
-from qiskit_dynamics import dispatch
-from qiskit_dynamics.dispatch import Array
+from qiskit_dynamics.array import Array
 
 
 class Signal:
@@ -106,7 +105,7 @@ class Signal:
             else:
                 self._envelope = lambda t: envelope * np.ones_like(t)
         elif callable(envelope):
-            if dispatch.default_backend() == "jax":
+            if Array.default_backend() == "jax":
                 self._envelope = lambda t: Array(envelope(t))
             else:
                 self._envelope = envelope
@@ -172,11 +171,9 @@ class Signal:
             return str(self.name)
 
         if self.is_constant:
-            return "Constant({})".format(str(self(0.0)))
+            return f"Constant({str(self(0.0))})"
 
-        return "Signal(carrier_freq={freq}, phase={phase})".format(
-            freq=str(self.carrier_freq), phase=str(self.phase)
-        )
+        return f"Signal(carrier_freq={str(self.carrier_freq)}, phase={str(self.phase)})"
 
     def __add__(self, other: "Signal") -> "SignalSum":
         return signal_add(self, other)
@@ -274,7 +271,8 @@ class DiscreteSignal(Signal):
 
     The envelope is specified by an array of samples ``s = [s_0, ..., s_k]``, sample width ``dt``,
     and a start time ``t_0``, with the envelope being evaluated as
-    :math:`f(t) =` ``s[floor((t - t0)/dt)]``.
+    :math:`f(t) =` ``s[floor((t - t0)/dt)]`` if ``t`` is in the interval with endpoints
+    ``start_time`` and ``start_time + dt * len(samples)``, and ``0.0`` otherwise.
     By default a :class:`~qiskit_dynamics.signals.DiscreteSignal` is defined to start at
     :math:`t=0` but a custom start time can be set via the ``start_time`` kwarg.
     """
@@ -301,30 +299,39 @@ class DiscreteSignal(Signal):
             name: name of the signal.
         """
         self._dt = dt
-        self._samples = Array(samples)
+
+        samples = Array(samples)
+
+        if len(samples) == 0:
+            zero_pad = np.array([0])
+        else:
+            zero_pad = np.expand_dims(np.zeros_like(Array(samples[0])), 0)
+        self._padded_samples = np.append(samples, zero_pad, axis=0)
+
         self._start_time = start_time
 
         # define internal envelope function
-        if self._samples.backend == "jax":
+        if samples.backend == "jax":
 
             def envelope(t):
                 t = Array(t).data
                 idx = jnp.clip(
                     jnp.array((t - self._start_time) // self._dt, dtype=int),
-                    0,
-                    len(self._samples) - 1,
+                    -1,
+                    len(self.samples),
                 )
-                return self._samples[idx]
+                return self._padded_samples[idx]
 
         else:
 
             def envelope(t):
+                t = Array(t).data
                 idx = np.clip(
                     np.array((t - self._start_time) // self._dt, dtype=int),
-                    0,
-                    len(self._samples) - 1,
+                    -1,
+                    len(self.samples),
                 )
-                return self._samples[idx]
+                return self._padded_samples[idx]
 
         Signal.__init__(self, envelope=envelope, carrier_freq=carrier_freq, phase=phase, name=name)
 
@@ -389,7 +396,7 @@ class DiscreteSignal(Signal):
         Returns:
             duration: The duration of the signal in samples.
         """
-        return len(self._samples)
+        return len(self.samples)
 
     @property
     def dt(self) -> float:
@@ -405,7 +412,7 @@ class DiscreteSignal(Signal):
         Returns:
             samples: the samples of the piecewise constant signal.
         """
-        return self._samples
+        return Array(self._padded_samples[:-1])
 
     @property
     def start_time(self) -> float:
@@ -418,7 +425,7 @@ class DiscreteSignal(Signal):
     def conjugate(self):
         return self.__class__(
             dt=self._dt,
-            samples=np.conjugate(self._samples),
+            samples=np.conjugate(self.samples),
             start_time=self._start_time,
             carrier_freq=-self.carrier_freq,
             phase=-self.phase,
@@ -426,36 +433,41 @@ class DiscreteSignal(Signal):
 
     def add_samples(self, start_sample: int, samples: List):
         """
-        Appends samples to the pulse starting at start_sample.
-        If start_sample is larger than the number of samples currently
-        in the signal the signal is padded with zeros.
+        Appends samples to the pulse starting at start_sample, filling any gap with zeros.
 
         Args:
-            start_sample: number of the sample at which the new samples
-                should be appended.
-            samples: list of samples to append.
+            start_sample: Index of the sample at which the new samples should be appended.
+            samples: List of samples to append.
 
         Raises:
-            QiskitError: if start_sample is invalid.
+            QiskitError: If start_sample is less than the current length of samples.
         """
-        if start_sample < len(self._samples):
-            raise QiskitError()
+        samples = Array(samples)
 
-        if len(self._samples) < start_sample:
-            self._samples = np.append(
-                self._samples, np.zeros(start_sample - len(self._samples), dtype=complex)
+        if len(samples) < 1:
+            return
+
+        if start_sample < len(self.samples):
+            raise QiskitError("Samples can only be added afer the last sample.")
+
+        zero_pad = np.expand_dims(np.zeros_like(Array(samples[0])), 0)
+
+        new_samples = self.samples
+        if len(self.samples) < start_sample:
+            new_samples = np.append(
+                new_samples, np.repeat(zero_pad, start_sample - len(self.samples))
             )
 
-        self._samples = np.append(self._samples, samples)
+        new_samples = np.append(new_samples, samples)
+        self._padded_samples = np.append(new_samples, zero_pad, axis=0)
 
     def __str__(self) -> str:
         """Return string representation."""
         if self.name is not None:
             return str(self.name)
 
-        return "DiscreteSignal(dt={dt}, carrier_freq={freq}, phase={phase})".format(
-            dt=self.dt, freq=str(self.carrier_freq), phase=str(self.phase)
-        )
+        return f"""DiscreteSignal(dt={self.dt}, carrier_freq={str(self.carrier_freq)},
+                phase={str(self.phase)})"""
 
 
 class SignalCollection:
@@ -569,7 +581,7 @@ class SignalSum(SignalCollection, Signal):
         SignalCollection.__init__(self, components)
 
         # set up routine for evaluating envelopes if jax
-        if dispatch.default_backend() == "jax":
+        if Array.default_backend() == "jax":
             jax_arraylist_eval = array_funclist_evaluate([sig.envelope for sig in self.components])
 
             def envelope(t):
@@ -594,7 +606,7 @@ class SignalSum(SignalCollection, Signal):
 
     def complex_value(self, t: Union[float, np.array, Array]) -> Union[complex, np.array, Array]:
         """Return the sum of the complex values of each component."""
-        if dispatch.default_backend() == "jax":
+        if Array.default_backend() == "jax":
             t = Array(t)
         exp_phases = np.exp(np.expand_dims(t, -1) * self._carrier_arg + self._phase_arg)
         return np.sum(self.envelope(t) * exp_phases, axis=-1)
@@ -608,7 +620,7 @@ class SignalSum(SignalCollection, Signal):
 
         default_str = str(self[0])
         for sig in self.components[1:]:
-            default_str += " + {}".format(str(sig))
+            default_str += f" + {str(sig)}"
 
         return default_str
 
@@ -758,7 +770,7 @@ class DiscreteSignalSum(DiscreteSignal, SignalSum):
 
         default_str = str(self[0])
         for sig in self.components[1:]:
-            default_str += " + {}".format(str(sig))
+            default_str += f" + {str(sig)}"
 
         return default_str
 
@@ -812,7 +824,7 @@ class SignalList(SignalCollection):
         super().__init__(signal_list)
 
         # setup complex value and full signal evaluation
-        if dispatch.default_backend() == "jax":
+        if Array.default_backend() == "jax":
             self._eval_complex_value = array_funclist_evaluate(
                 [sig.complex_value for sig in self.components]
             )
